#!/usr/bin/python

# Lost Cities
# Copyright 2003, David "Novalis" Turner <novalis@fsf.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from random import *
import asyncore
import socket
import pickle

from Connection import Connection
from common.Utility import *

class Server (asyncore.dispatcher):
    def __init__ (self):
        asyncore.dispatcher.__init__(self)
        self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr ()
        self.bind (('127.0.0.1', 10611))
        self.listen (5)
        self.games = {}
        self.passwords = self.loadPlayerDB ()
        self.connections = {}
        self.challenges = {}

    def loadPlayerDB (self):
        try:
            f = open ("players.p", "r")
        except:
            return {}
        return pickle.load (f)

    def savePlayerDB (self):
        try:
            f = open ("players.p", "w")
        except:
            return
        pickle.dump (self.passwords, f)

    def handle_accept (self):
        Connection (self.accept(), self)

    def closeUserConn (self, username):
        #fixme: tell other player
        #this seems to get called more than once per close so:
        if self.connections.has_key (username):
            del self.connections [username]
            del self.challenges [username]

    def createAccount (self, username, password, connection):
        if self.passwords.has_key (username):
            return 0
        
        self.passwords [username] = password

        self.doLogin (username, connection)
        return 1

    def getLobbyPlayers (self):
        lobbyPlayers = []
        for (name, conn) in self.connections.iteritems ():
            if conn.mode == 'lobby':        
                lobbyPlayers.append (name)
                
        return lobbyPlayers
                
    def doLogin (self, username, connection):
        self.connections [username] = connection
        self.challenges [username] = []

        connection.sendl ('loggedIn', 'dummy')
        
        lobbyPlayers = self.getLobbyPlayers ()

        for player in lobbyPlayers:
            self.connections [player].sendl ('user', username)

        if lobbyPlayers:
            connection.sendl ('user', '\0'.join (lobbyPlayers))

    def chat (self, user, text):
        for player in self.getLobbyPlayers ():
            self.connections [player].sendl ('chat', user + "\0" + text)
        
    def login (self, username, password, connection):
        if not self.passwords.has_key (username) or self.passwords [username] != password:
            return 0
        if self.connections.has_key (username):
            return 0 #already logged in
        
        self.doLogin (username, connection)
        
        return 1

    def challenge (self, username, opponent):

        if not opponent in self.connections:
            return 0
        
        if opponent in self.challenges [username]:
            pass #already challenged, fixme: unchallenge

        if username in self.challenges [opponent]:
            #clear out the challenges
            self.challenges [username] = []
            self.challenges [opponent] = []
            
            self.connections [username].playing (opponent)
            self.connections [opponent].playing (username)
            
            #they have already challenged us, start the game
            game = Game (Player (self, self.connections [username]),
                         Player (self, self.connections [opponent]))
            
        else:
            #issue the challenge
            self.challenges [username].append (opponent)
            self.connections [opponent].sendl ("challenge", username)

        return 1           

class Player:
    def __init__ (self, server, conn):
        self.conn = conn
        conn.setPlayer (self)
        self.server = server
        self.hand = []
        self.opponent = None
        
    def startGame (self, game):
        self.game = game
        for i in range (0, 8):
            card = self.game.draw ()
            self.hand.append (card)
            self.sendl ("youdrew", card)
            
    def playCard (self, card):
        if not self.checkMode ('play'):
            return
        
        if not card in self.hand:
            self.error ("notInHand")
            return
        
        if not self.game.playCard (card):
            self.error ("playOutOfOrder")
        
        self.opponentPlayer.sendl ("played", card)

        self.game.mode = 'draw'

    def discard (self, card):

        if not self.checkMode ('play'):
            return
        
        if not card in self.hand:
            self.error ("notInHand")
            return

        self.game.discard (card)
        self.opponentPlayer.sendl ("discarded", card)

        self.game.mode = 'draw'
        self.hand.remove (card)

    def checkMode (self, mode):
        """Ensure that the game is in the mode passed in the mode var, and
        if not, send an error message to the client"""
        if not self.game.turn == self:
            self.error ("outOfTurn")
            return 0
        if not self.game.mode == mode:
            self.error ("alreadyPlayed")
            return 0
        return 1

    def drawDeck (self):
        """Draw from the deck """
        if not self.checkMode ('draw'):
            return

        card = self.game.draw ()
        self.hand.append (card)
        self.sendl ("youdrew", card)
        self.opponentPlayer.sendl ("theydrew deck")
        if len (self.game.deck) == 0:
            for x in (self, self.opponentPlayer):
                x.sendl ("gameOver", "morx")
        else:
            self.game.doneTurn ()

    def drawDiscard (self, suit):
        if not self.checkMode ('draw'):
            return

        card = self.game.discardDraw (int (suit))
        if card == -1:
            self.error ("emptyDiscardPile")
            return
        self.hand.append (card)
        self.opponentPlayer.sendl ("theydrew", suit)
        self.game.doneTurn ()
        
    def turn (self):
        self.sendl ("turn dummyArg")

    def notTurn (self):
        self.sendl ("notturn dummyArg")

    def error (self, err):
        self.sendl ("error", err)

    def sendl (self, *args):
        self.conn.sendl (*args)

class Game:
    def __init__ (self, player1, player2):
        self.players = [player1, player2]
        player1.opponentPlayer = player2
        player2.opponentPlayer = player1
        self.playersToIdx = { player1 : 0, player2 : 1 }

        self.deck = range (0, 5*13)
        #remove the 1 cards, since the deck is only 2-10
        self.deck = filter (lambda x: x % 13 != 3, self.deck)
        shuffle (self.deck)

        self.discards = []
        self.played = [[],[]]
        for i in range (0, 5):
            self.discards.append ([])
            self.played [0].append ([])
            self.played [1].append ([])

        player1.startGame (self)
        player2.startGame (self)

        if random () > 0.5:
            self.turn = player1
            player2.notTurn ()
        else:
            self.turn = player2
            player1.notTurn ()

        self.turn.turn ()
        self.mode = 'play'
        
    def playCard (self, card):
        suit = int (card / 13)
        rank = card % 13
        played = self.played [self.playersToIdx [self.turn]]
        if canPlayCard (suit, rank, played):
            played [suit].append (card)
            return 1
        else:
            return 0

    def discard (self, card):
        suit = int (card / 13)
        rank = card % 13
        self.discards [suit].append (card)

    def doneTurn (self):
        self.turn = self.players [self.playersToIdx [self.turn] ^ 1]
        self.mode = 'play'

    def draw (self):
        return self.deck.pop ()

    def discardDraw (self, suit):
        if len (self.discards [suit]) == 0:
            #nothing discarded yet
            return -1

        return self.discards [suit].pop ()
        
